import libfmp4
from manifest_edit.pipeline import Pipeline
from manifest_edit.context import Context


def _isMemberFromLibfmp4(member):
    # This is the best way I found to accomplish the task. I can't tell if it
    # is very robust.
    try:
        # sometimes __module__ is present and it is None (method-wrapper
        # type)
        attrname = getattr(member, "__module__", '')
        # this instruction is surely not robust, if libfmp4 is imported
        # with "as" under another name, so it is important, if we want to use
        # this approach, that this code is executed in a controlled module
        return attrname and attrname.startswith('libfmp4')
    except Exception as e:
        Context.log_error('Exception {e} while trying to access "__module__" on \
            member {member}'.format(e=e, member=member))
        return False


def _isMemberAList(member):
    # Even empty lists should be substituted, in case someone wants to append
    # to it.
    # i.e. manifest._utc_timings when empty is just an empty python list
    # and there is no way (that I found) to infer the data type it is
    # supposed to hold. In doubt, you will need to consider this a candidate
    # for "copying".
    return isinstance(member, list)


def _membersShortList(type_in):
    return [member_name for member_name in dir(type_in)
            if not member_name.startswith('__')]


def _membersToBeProcessed(type_in):
    # The idea is to get the shortest possible list of members that we
    # should recurse into, avoiding the obvious __init__, __ge__ and so on
    # and selecting only lists or types belonging to libfmp4
    return [member_name for member_name in _membersShortList(type_in)
            if (
                _isMemberAList(getattr(type_in, member_name)) or
                _isMemberFromLibfmp4(getattr(type_in, member_name))
    )]


def _copyStdVectors(type_in):
    '''
    The idea is to identify every member whose name starts with an underscore
    and add a dynamic attribute with the same name without underscore, i.e.:

    manifest._periods ==> manifest.periods will be added

    We need to handle any possible level of nesting of course, so this method
    will recurse the received type_in.

    The types are a bit peculiar here. We cannot use the standard "inspect"
    method of python because the typical manifest_t data types and related
    do not have all the metadata normally associated with native python types
    such as "__dict__". But we can rely on "__dir__".

    We will use the fact that manifest_t only used stl containers are list
    and will limit recursive iteration to such containers.

    Recursion will stop when all data types of a given member do not belong
    to libfmp4 anymore (i.e. are native C++ types).
    '''

    # Iterate over a subset of members that may need to be copied
    for member_name in _membersToBeProcessed(type_in):
        member = getattr(type_in, member_name)
        # if the member is a list, recurse on each element
        if isinstance(member, list):
            for item in member:
                Context.log_debug("_copyStdVectors recursing into {} in {}".format(
                    type(item), type(member)))
                _copyStdVectors(item)
        else:
            # otherwise just recurse on it
            Context.log_debug(
                "_copyStdVectors recursing into {}".format(type(member)))
            _copyStdVectors(member)

        # if the member name starts with an underscore, then create a copy
        # i.e.: "instance._member" will have an "instance.member" copy,
        # obtaining at all effects an editable copy of the original data type
        if member_name.startswith('_') and not member_name.startswith('__'):
            # Ok, for the moment we are using this but this is very unsafe.
            # The newly created list is indeed a native python list, but
            # as such it does not have any type hint. User can add
            # whatever type they want and the operation will not fail.
            # But it definitely will screw up things when we try and copy
            # it back in a std::vector which is strictly typed.
            Context.log_debug("_copyStdVectors copying {} to {}".format(
                member_name, member_name[1:]))
            setattr(type_in, member_name[1:], getattr(type_in, member_name))


def _membersToBeRemoved(type_in):
    # All the types that have been dynamically added went into __dict__
    # This is handy.
    return [member_name for member_name in getattr(type_in, "__dict__", {})]


def _removePythonMembers(type_in):
    '''
    Removes every member present in class' python __dict__.
    Works only if the only containers are list.
    '''
    to_be_removed = _membersToBeRemoved(type_in)
    if len(to_be_removed):
        for member_name in to_be_removed:
            member = getattr(type_in, member_name)
            # if the item is a list, recurse and substitute. Only list are
            # supported so far.
            if isinstance(member, list):
                for item in getattr(type_in, member_name):
                    Context.log_debug("_removePythonMembers recursing into {} in {}".format(
                        type(item), type(member)))
                    _removePythonMembers(item)
            else:
                Context.log_debug(
                    "_removePythonMembers recursing into {}".format(type(member)))
                _removePythonMembers(member)

            # copy back and remove the attribute
            # This can blow up if containers do not have the same type
            # expected by the std::vector
            Context.log_debug("_removePythonMembers copying {} to {}".format(
                member_name, "_"+member_name))
            setattr(type_in, "_"+member_name, getattr(type_in, member_name))
            Context.log_debug(
                "_removePythonMembers deleting {}".format(member_name))
            delattr(type_in, member_name)


# These operations need to be executed once only (at startup time). Only after
# executing this method you are ready to receive manifests for editing
def setup(extension, mp4process_context):
    Context.mp4context = mp4process_context
    _setup(extension, mp4process_context.pipeline_filename)


# This version of setup is only intended for unit tests, where an mp4context
# is not available
def _setup(extension, pipeline_filename):
    if extension not in Pipeline.SUPPORTED_MANIFEST_FORMATS:
        # in case something goes wrong during setup, update_manifest
        # should just do nothing
        raise Exception("Unsupported extension {}".format(extension))

    Context.log_debug("Using libfmp4 = {}".format(libfmp4.version_string()))

    # Context variables, need to be initialized at startup
    Context.format = extension
    Context.pipelines = Pipeline.build_pipeline(pipeline_filename)

# this is the entry point for updating an MPEG DASH manifest
# def update_manifest(manifest: libfmp4.mpd.manifest) -> libfmp4.mpd.manifest:


def update_manifest(manifest: libfmp4.mpd.manifest) -> libfmp4.mpd.manifest:
    Context.log_debug("entering update manifest")
    # This call will dynamically add native python list members for every
    # std::vector in the manifest

    Context.log_debug("Copying STL containers")
    _copyStdVectors(manifest)

    Context.log_debug("Pipeline processing...")
    # Actual pipeline processing
    Context.pipelines[Context.format].run(input_manifest=manifest)

    # This call will copy python lists back into the original stl containers
    # and will remove dynamically added members
    Context.log_debug("Removing copied STL containers")
    _removePythonMembers(manifest)

    Context.log_debug("exiting update manifest")

    return manifest
