from schema import Schema, Optional, Or, And
from ..context import Context


class ManifestIterator(object):

    _recurse_keys = ['periods', 'adaptation_sets', 'representations']

    _schema = Schema(
        {
            'periods': Or(dict, [
                And({
                    str: Or(str, int),
                    Optional('adaptation_sets'): Or(dict, [
                        And({
                            str: Or(str, int),
                            Optional('representations'): Or(dict, [
                                lambda x: len(x.keys()) > 1
                            ])
                        }, lambda x: len(x.keys()) > 1)
                    ])
                }, lambda x: len(x.keys()) > 1)
            ])
        }
    )

    def __init__(self, configuration):
        self.config = configuration
        self._is_valid = self._schema.is_valid(self.config)

    def validate(self):
        return self._is_valid

    def __call__(self, root_node, process_keys, config=None):
        '''
        This method will expect a nested configuration dictionary that the
        user would have configured to express which particular list or lists
        (in the many available in the libfmp4 manifest_t type), he wants to
        process, and how. [config argument]

        This method will expect also a list of "process_keys", which
        the developer will have chosen to express the parameters of the
        particular manifest manipulation he needs to perform on the element
        or on the list. These will be used as a "sentinel" to stop recursion
        on the config dictionary and yield the corresponding manifest item or list.

        Depending on how the configuration file is written, this method
        will yield entire lists or single elements. It is responsibility
        of the developer to expect the right return type based on the
        cfg file he wrote.

        The expected structure is that the configuration
        dictionary contains a key nesting which is:

        - periods (level 0) OR adaptation_sets (level 1) OR representations
        (level 2)

        Notice that level 0 is the only mandatory one.

        The corresponding value would be either of:

        - a list of dictionaries, where each item of the list will contain a
        selection key (i.e. id : "1", id : "*") and either of:
            - the process_key dictionary:
                in this case individuals item satisfying the selection will be
                yield
            - one of the possible recurse keys [periods, adaptation_sets,
            representations]
                in this case the entire processing will restart from the next
                level
        - the process_key dictionary:
            in this case the entire list will be yield

        Example 1 of a valid config:
        periods:
            - id : "*"
            adaptation_sets:
                - contentType :  "video"
                representations:
                    orderBy :      "bandwidth"
                    orderCriteria: "desc"
                - contentType :  "audio"
                representations:
                    orderBy :      "bandwidth"
                    orderCriteria: "desc"

        that would be passed to this method as:

            manifestIterator(
                manifest, config, ['orderBy', 'orderCriteria']
                )

        The calling method would receive "representations" lists, taken from
        "any" periods and from adaptation_sets having type "video" or "audio",
        along with the {'orderBy': 'bandwidth, 'orderCriteria': 'desc'}
        dictionary.

        Example 2 of a valid config:
        periods:
            - id : "*"
            adaptation_sets:
                - contentType :  "video"
                do :      "this"
                - contentType :  "audio"
                do :      "that"

        that would be passed to this method as:

            manifestIterator(
                manifest, config, ['do']
                )

        The calling method would receive individual adaptatation_set_t objects
        (not the entire "adaptation_sets" lists), taken from
        "any" periods and having contentType "video" or "audio". Each of those
        will be yielded with the corresponding {'do' :'this'} or {'do' :'that'}
        dictionary.
        '''
        if not self._is_valid:
            yield None, None

        if not config:
            config = self.config

        for node_name, node_config in config.items():

            # This is periods, adaptation_sets or representations
            node = getattr(root_node, node_name)

            # This happens when you are in a "leaf" of a configuration that
            #  will return an entire list, not each individual elements.
            # That is, something like this:
            #
            #      periods:
            #       do :      "this"
            #
            # is going to yield the entire manifest.periods list
            if isinstance(node_config, dict):
                if all(key in process_keys for key in node_config.keys()):
                    # No selection key, just yield everything in list form
                    # Covered by test_period_selection_wildcard and many
                    #  others
                    yield node_config, node
                else:
                    # This can only happen if you are trying to find
                    #  configuration keys that are not present in the config
                    #  file. Could happen.
                    Context.log_debug(
                      'The provided process keys {p_ks} \
                       do not match config keys {c_ks}'.format(
                           p_ks=process_keys,
                           c_ks=node_config.keys()
                           ))

            # When instead you have a configuration that has some sort of
            #  selection items, individual elements will be returned
            # That is, something like this:
            #
            #      periods:
            #      - '*' : '*'
            #        do :      "this"
            #
            # is going to yield each individual period from the
            # manifest.periods list
            elif isinstance(node_config, list):
                for item in node_config:
                    # You can either find one of the processing keys
                    # OR a recurse_key (which we ignore for now)
                    # OR a field on which to perform node selection

                    # Keys that can be used to perform selection on this node.
                    item_selection_keys = [
                        key for key in item.keys() if key not in self._recurse_keys+process_keys]

                    # If this node will be selected, you will have to recurse in
                    #  these elements too
                    node_recurse_keys = [
                        key for key in item.keys() if key in self._recurse_keys]
                    # unless you are in a leaf
                    node_recurse_key = node_recurse_keys[0] if len(
                        node_recurse_keys) else None
                    # Iterate over the manifest node (periods, adaptation_sets or
                    #  representations) and look for items with the configured
                    #  property and value
                    for node_item in node:
                        # Notice that if the configuration specifies a property
                        # that does not exist in the manifest item, then this will
                        # not be selected, even if its configured value is '*'.
                        # If instead the property name is '*', then it will
                        # always be selected.
                        if self.node_matches(item_selection_keys, node_item, item):
                            if node_recurse_key:
                                # Covered by test_period_selection_by_value and
                                # test_adaptation_set_selection_by_value
                                # Covered by test_adaptation_set_selection_wildcard
                                yield from self(node_item, process_keys, {node_recurse_key: item[node_recurse_key]})
                            else:
                                # Covered by test_selection_by_value_on_leaf
                                # Covered by test_adaptation_set_selection_wildcard_on_leaf
                                yield item, node_item
            # else:
                # No other checks on node_config are necessary if config
                #  validation works correctly

    def node_matches(self, item_selection_keys, node_item, item):
        # This function will return True if this manifest node_item matches 
        #  the key, attribute pairs present in the the configuration "item"
        #
        matches_list = [self.node_matches_one(key, node_item, item) for key in item_selection_keys]
        return all(matches_list)

    def node_matches_one(self, item_selection_key, node_item, item):
        # We have a match if:
        # - the configured key is the wildcard value '*'
        # - the configured key exists in the manifest and its value is
        #   either matching the one specified in the config or the
        #   latter is '*'

        # Notice that this piece of code hides a very strong assumption, that
        #  seems to be used everywhere in libfmp4: any optional field
        #  is always present in manifest_t but has a default "illegal" value
        #  that is:
        #  - 0 for numeric fields
        #  - empty strings for string fields
        # We are going to leverage the fact that these values all evaluate to
        #  false-ish in python and just test for "node_value"

        # Quickly return in case of wildcard on key
        if item_selection_key == '*':
            return True

        # Quickly return False if the selection key is not present in the
        # manifest ==> the value is false-ish
        node_value = getattr(
            node_item, item_selection_key, None)
        if not node_value:
            return False

        # If the wildcard value was selected, we can just return True
        value_read_from_config_file = item[item_selection_key]
        if value_read_from_config_file == '*':
            return True

        #Else we need to compare the value in the yml file with the one in the
        # manifest. We need to make sure we are comparing the same types or
        # weird things can happen
        node_value_type = type(node_value)
        config_value = node_value_type(value_read_from_config_file)
        return node_value == config_value
