from manifest_edit import BasePlugin
from manifest_edit import ManifestIterator
import re
from ..context import Context


class Plugin(BasePlugin):
    '''
    manifest_order plugin.

    The purpose of this plugin is to rearrange track order in manifest files
    based on several possible criteria.

    Limitations: only implemented for .mpd manifests

    It expects to find some keywords in the configuration files to specify
    the sorting action to be performed:
    - 'orderBy' is to perform a full sorting by a key
    - 'manualOrder' is to perform "manual" sorting, assigning a position to
       elements based on some of their properties
    '''

    def reorderManifestLists(self, manifest):

        manifest_iterator = ManifestIterator(self.config)

        # We will just use manifest_iterator with all possible keywords. It will
        # return nothing in case the keyword is not present in self.config .
        # Maybe this can be optimized and only use it on what we know it's
        # present.
        for order_config, list_to_order in manifest_iterator(manifest,
                                                             ['orderBy', 'orderCriteria']):
            list_to_order.sort(
                key=lambda repr: getattr(
                    repr, order_config['orderBy'], ''),
                reverse=True if order_config['orderCriteria'] == 'desc' else False
            )

        for order_config, list_to_order in manifest_iterator(manifest,
                                                             ['manualOrder']):

            # Just checking the user has not forget the ordering list
            if not order_config['manualOrder']:
                Context.log_warning(
                    'It seems like a manualOrder was specified but no rules \
                        for ordering have been provided in the .yml config \
                        file. Reordering cannot be performed!'
                    )

                return

            # iterate on list of key,value and take note of index of list
            for idx, config_item in enumerate(order_config['manualOrder']):
                # get from config the key to look for
                # get from config the re to match the value
                for key, expression in config_item.items():
                    # See if an element of the manifest matches the expression
                    # and, if yes, get its index in the list it belongs
                    try:
                        regex = re.compile(expression)
                    except re.error:
                        Context.log_warning(
                        'The provided expression {} does not seems a valid \
                            python regular expression and will be ignored'
                            .format(expression)
                        )
                        continue

                    found_at = [i for i, x in enumerate(list_to_order) if regex.search(
                        str(getattr(x, key, None)))]

                    if len(found_at):
                        # if a matching element is found, we will move it to the
                        # desired location. If it is found multiple times, only
                        # the firts occurence is considered. If it is not found,
                        # nothing happens
                        list_to_order.insert(
                            idx, list_to_order.pop(found_at[0]))

                    # We now just break out of the config_item.items iteration
                    # because we don't support, for the moment, matching items
                    # by multiple key,expression couples.
                    break

    def mpd(self, manifest):
        self.reorderManifestLists(manifest)
