from manifest_edit import BasePlugin
from manifest_edit import UTCTiming
from manifest_edit import Context


class Plugin(BasePlugin, UTCTiming):
    '''
    utc_remove plugin.

    The purpose of this plugin is to remove the "UTCTiming" element from a
    manifest, if present.

    For details on the "UTCTiming" element, see ISO/IEC 23009-1 5.3.1.2,
    5.3.1.3, 5.8.4.11 and 5.8.5.7.

    Several UTC Timing Scheme are possible, each requiring its own syntax for
    the value (see 5.8.5.7).

    You can choose to remove only those UTCTiming elements having a specific
    Scheme and Value by means of the .yml configuration file.

    "Scheme" can be specified using one of the following strings:

    - 'ntp'
    - 'sntp'
    - 'http-head'
    - 'http-xsdate'
    - 'http-iso'
    - 'http-ntp'
    - 'direct'

    Additionally, you can decide to just remove the UTCTiming element, whatever
    scheme it may have. This is done by specifying the value

    - '*'

    Example of configuration file removing only UTCTiming elements having the
    "urn:mpeg:dash:utc:ntp:2014" scheme:

    - mpd:
        - manifest_edit.plugins.utc_remove:
            - 'ntp':

    Example of configuration file removing any possible UTCTiming elements:
    - mpd:
        - manifest_edit.plugins.utc_remove:
            - '*':

    Limitations: only implemented for .mpd manifests
    '''

    def remove(self, manifest):
        # remove only works on manifests that do have an UTCTiming tag
        if not len(manifest.utc_timings):
            return
        # Not the most efficient way to do it but until we find a way to avoid
        # copying STL containers back and forth it's just the way it is
        for scheme_to_remove in self.config:
            if scheme_to_remove == '*':
                manifest.utc_timings.clear()
                break
            elif scheme_to_remove in self.SUPPORTED_UTC_SCHEMES.keys():
                scheme_id_uri_to_remove = self.SUPPORTED_UTC_SCHEMES[
                    scheme_to_remove
                ]
                for descriptor in manifest.utc_timings:
                    manifest.utc_timings[:] = [
                        timing for timing in manifest.utc_timings if
                        descriptor.scheme_id_uri != scheme_id_uri_to_remove
                    ]
            else:
                Context.log_error(
                    'Unsupported scheme {scheme}'.format(scheme=scheme_to_remove))

    def mpd(self, manifest):
        self.remove(manifest)
